#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 19 17:25:37 2022

exercise 2.c solving the lorenz system

@author: richter
"""

import numpy as np
import matplotlib.pyplot as plt
import torch
import time



## The right hand side of the ODE:  x'(t) = A x(t)
def A(x):
    return np.array([-10.0 * x[0]+10.0*x[1],
                     28.0*x[0]-x[1]-x[0]*x[2],
                     x[0]*x[1]-8./3.*x[2]])

## One step of Heun's 2nd order method
def heun(x,h):
    return x+0.5*h*A(x) + 0.5*h*A(x+h*A(x))



## Create random training data
Ntrain = 1000
x = (np.array([-15,-20,10])+np.array([30,40,30])*np.random.rand(Ntrain,3)).astype(np.float32)  # Make sure that it is float!


## for each of these points create the y-value, that is the
## solution after htrain = 0.1
print('Creating test data for {0} random initial points'.format(Ntrain))
y = x.copy() # first, we set y as the initial value
htrain = 0.1   # time step for neural network prediction
hheun  = 0.01 # time step for Heun's method 
nheun = int(htrain/hheun+1.e-8)  # number of Heun-steps per network step
assert  abs(hheun*nheun-htrain)<1.e-10,'Test data is constructed with wrong time step'

## Determine the end-points by solving Ntrain - times the Lorenz System
for n in range(Ntrain):
    for i in range(nheun):          # time-stepping
        y[n,:] = heun(y[n,:],hheun) # we overwrite y[n,:] since we only need the end-point

## transform to pytorch
xtrain = torch.from_numpy(x)
ytrain = torch.from_numpy(y)
    
## The neural network
## We create the model
model = torch.nn.Sequential(
                torch.nn.Linear(3,10),  # 1 input -> 2
                torch.nn.Sigmoid(),    # sigmoid-activation
                torch.nn.Linear(10,10),  # 1 input -> 2
                torch.nn.Sigmoid(),    # sigmoid-activation
                torch.nn.Linear(10,3)   # 2 -> 1
                )                      # no further activation

learningrate, maxiter =0.001,100000
loss_fn = torch.nn.MSELoss(reduction='sum')
optimizer = torch.optim.Adam(model.parameters(),lr=learningrate,weight_decay=0)


for t in range(maxiter+1):   # optimization loop
    torch_prediction = model(xtrain)        # evaluate the model
    loss = loss_fn(torch_prediction,ytrain) # compute the loss
    if t%1000==0:                     # output
        print(t,loss.item()/np.sqrt(Ntrain),sep='\t')
    optimizer.zero_grad()                   # 
    loss.backward()                         # compute the gradient
    optimizer.step()                        # one step



